import * as toml from "toml";
import { CurseForgeClient } from "curseforge-api";
import Resource from "./resource";

// const CURSE_CLIENT = new CurseForgeClient();

export interface Metafile {
  name: string;
  filename: string;
  side: "server" | "client" | "both";
  provider: Provider;
}

export abstract class Provider {
  constructor(
    public hash: string,
    public hashFormat: string,
  ) {}

  abstract get downloadUrl(): Resource;
}

export class UrlProvider extends Provider {
  constructor(
    public hash: string,
    public hashFormat: string,
    private url: string,
  ) {
    super(hash, hashFormat);
  }

  get downloadUrl(): Resource {
    return new Resource(this.url);
  }
}

export class ModrinthProvider extends UrlProvider {
  constructor(
    hash: string,
    hashFormat: string,
    url: string,
    public modId: string,
    public versionId: string,
  ) {
    super(hash, hashFormat, url);
  }
}

export class GitHubProvider extends UrlProvider {
  constructor(
    hash: string,
    hashFormat: string,
    url: string,
    public branch: string,
    public regex: string,
    public slug: string,
    public tag: string,
  ) {
    super(hash, hashFormat, url);
  }
}

export class CurseForgeProvider extends Provider {
  constructor(
    hash: string,
    hashFormat: string,
    public mode: string,
    public fileId: number,
    public projectId: number,
  ) {
    super(hash, hashFormat);
  }

  get downloadUrl(): Resource {
    return new Resource(
      `https://www.curseforge.com/api/v1/mods/${this.projectId}/files/${this.fileId}/download`,
    );
  }
}

/**
 * Represents a file entry in the Packwiz index.
 */
export class IndexFileEntry {
  readonly file: Resource;
  readonly hash: string;
  readonly metafile: boolean;

  constructor(
    file: string | Resource,
    hash: string,
    metafile: boolean = false,
  ) {
    this.file = file instanceof Resource ? file : new Resource(file);
    this.hash = hash;
    this.metafile = metafile;
  }

  async parse(): Promise<Metafile> {
    if (this.metafile === false) {
      throw new Error(
        "Cannot parse non-metafiles! Use `file.readText()` instead to get the file contents.",
      );
    }
    const fileContent = await this.file.fetchContents();
    const parsed = toml.parse(fileContent);

    const metafile: Metafile = {
      name: parsed.name,
      filename: parsed.filename,
      side: parsed.side,
      provider: this.parseProvider(parsed),
    };

    return metafile;
  }

  private parseProvider(parsed: any): Provider {
    if (parsed.update?.modrinth) {
      return new ModrinthProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        parsed.download.url,
        parsed.update.modrinth["mod-id"],
        parsed.update.modrinth["version"],
      );
    } else if (parsed.update?.curseforge) {
      return new CurseForgeProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        parsed.download.mode,
        parsed.update.curseforge["file-id"],
        parsed.update.curseforge["project-id"],
      );
    } else if (parsed.update?.github) {
      return new GitHubProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        parsed.download.url,
        parsed.update.github.branch,
        parsed.update.github.regex,
        parsed.update.github.slug,
        parsed.update.github.tag,
      );
    } else if (parsed.download) {
      return new UrlProvider(
        parsed.download.hash,
        parsed.download["hash-format"],
        parsed.download.url,
      );
    } else {
      throw new Error("Unknown provider in TOML.");
    }
  }
}

/**
 * Represents the structure of a Packwiz index file, as well as providing its location.
 */
export interface PackwizIndex {
  location: Resource;
  hashFormat: string;
  files: IndexFileEntry[];
}

// TODO: parse pack.toml too instead of just the index file
// This is because pack.toml contains important information like modloader versions

/**
 * Parses a Packwiz index file and returns its contents.
 *
 * @param indexFilePath - The path of the TOML file.
 * @returns The parsed index as a structured object.
 */
export async function parsePackwizIndex(
  indexFilePath: string,
): Promise<PackwizIndex> {
  const indexFile = new Resource(indexFilePath);
  const parsed = toml.parse(await indexFile.fetchContents());
  return {
    location: indexFile,
    hashFormat: parsed["hash-format"],
    files: parsed.files.map(
      (file: any) =>
        new IndexFileEntry(
          indexFile.parent.join(file.file),
          file.hash,
          file.metafile,
        ),
    ),
  };
}
